#!/bin/sh
#
# capcalc - Calculate capacity based on memory, CPU, or load.
#
# Copyright (c) 2014-2023 Occam's Razor. All rights reserved.
#
# See the LICENSE file distributed with this code for restrictions on its use
# and further distribution.
# Original distribution available at <http://www.occam.com/tools/>.
#
# $Id: capcalc.sh,v 1.2 2023/06/09 18:13:51 leonvs Exp $
#
# TODO
#	GPU functionality tested in only one environment; likely very fragile
#

PATH=/bin:/usr/bin

usage () {
	echo "usage:	$progname -h
	$progname { -c | -g | -l | -m } proc_name [proc_name...]

	-h: Prints this usage statement.
	-c: Calculates capacity based on CPU usage.
	-g: Calculates capacity based on GPU usage.
	-l: Calculates capacity based on load average.
	-m: Calculates capacity based on memory usage.
	'proc_name' is the name of the process to measure." >&2
	exit 2
}

setup () {
	progname=`basename $0`

	while getopts hcglm c; do
		case $c in
			h)	usage;;
			c)	mode="cpu";;
			g)	mode="gpu";;
			l)	mode="load";;
			m)	mode="mem";;
			\?)	usage;;
		esac
	done
	shift `expr $OPTIND - 1`

	# Need a mode set.
	[ -z $mode ] && usage

	# Need at least one argument.
	[ $# -ge 1 ] || usage

	procs=$*

	if [ $mode = "mem" ]; then
		# Determine a ratio between total memory used and total RSS of
		# all processes, to use in normalizing RSS. No, it's not very
		# accurate, but there is no accurate way to determine how much
		# memory a set of processes uses. Totaling up Private memory per
		# process, or PSS per process, results in numbers that are too
		# small; totaling up RSS results in numbers too big (due to
		# multi-counting shared pages). But this calculation seems like
		# a decent middle ground, and fairly safe for making scaling
		# decisions.
		#
		# These are in KB.
		rssAll=`ps auxww | grep -v ^USER | awk '{ print $6 }' | awk '{ total += $1}END{ print total}'`
		if [ "X`free | grep buffers/cache | wc -l`" = "X0" ]; then
			used=`free | grep Mem: | awk '{ print $3 }'`
		else
			used=`free | grep buffers/cache | awk '{ print $3 }'`
		fi
		total=`free | grep Mem: | awk '{ print $2 }'`
	else
		# Normalize by number of CPU cores.
		cores=`cat /proc/cpuinfo | grep processor | wc -l`
	fi
}

cpucalc () {
	cpu=`ps auxww | grep $proc | grep -v grep | grep -v $progname | awk '{ print $3 }' | awk '{ total+=$1 }END{ print total }'`
	# expr can't handle division of non-integer values, so use awk.
	cpuPer=`echo "$cpu $count" | awk '{ print $1/$2 }'`
	procMax=`echo "$cores $cpuPer" | awk '{ print $1*100/$2 }' 2>/dev/null`
	echo "$count $proc processes, cpu per proc $cpuPer, up to $procMax processes per host"
}

gpucalc () {
	gputool="nvidia-smi"
	#`which $gputool 1>/dev/null 2>/dev/null` || ( echo "$gputool not present" && exit )
	gpuCt=`$gputool 2>/dev/null | grep $proc | wc -l`
	[ "X$gpuCt" = "X0" ] && echo "$gpuCt $proc processes on GPU" && exit
	gpuUse=`$gputool | grep $proc | awk '{ print $8 }' | sed 's/MiB//' | awk '{ total+=$1 }END{ print total }'`
	gpuTot=`$gputool | grep MiB | awk '{ print $10 }' | grep MiB | sed 's/MiB//' | awk '{ total+=$1 }END{ print total }'`
	gpuPer=`echo "$gpuUse $gpuCt" | awk '{ print $1/$2 }'`
	procMax=`echo "$gpuTot $gpuPer" | awk '{ print $1/$2 }' 2>/dev/null`
	echo "$gpuCt $proc processes, gpu per proc $gpuPer MiB, up to $procMax processes per host"
}

loadcalc () {
	load=`uptime | cut -d, -f4 | cut -d: -f2`
	# expr can't handle division of non-integer values, so use awk.
	loadPer=`echo "$load $count" | awk '{ print $1/$2 }'`
	procMax=`echo "$cores $loadPer" | awk '{ print $1/$2 }' 2>/dev/null`
	echo "$count $proc processes, load per proc $loadPer, up to $procMax processes per host"
}

memcalc () {
	rssTot=`ps auxww | grep $proc | grep -v grep | grep -v $progname | awk '{ print $6 }' | awk '{total += $1}END{print total}'`
	# normalized RSS in MB
	rssNorm=`expr $rssTot \* $used / $rssAll / 1024`
	rssPer=`expr $rssNorm / $count`
	procMax=`expr $total / 1024 / $rssPer`
	#echo "normed RSS total $rssNorm MB"
	echo "$count $proc processes, normed RSS per proc $rssPer MB, up to $procMax processes per host"
}

setup $*
for proc in $procs; do
	# Yes, running ps each time wastes cycles and leads to possibly
	# inconsistent data since it isn't collected at a single point in time.
	# The right way to do this would be to get process data once and then
	# reference that for calculations. But this is good enough.
	# I'm so tempted to rewrite this in Perl.
	count=`ps auxww | grep $proc | grep -v grep | grep -v $progname | wc -l`
	[ "X$count" = "X0" ] && echo "$count $proc processes" && exit
	${mode}calc
done
